﻿using System;
using System.Linq;
using System.IO.Ports;
using System.Collections.Generic;
using System.Text.RegularExpressions;


/* KeiganMotor API
 * 
 * Receive Data from KeiganMotor
 * 
 */

public partial class KeiganMotor_USB
{

    public class InterfaceSetting
    {
        public byte rawValue;
        public bool isBLEEnabled;
        public bool isLinkageEnabled;
        public bool isRadioEnabled;
        public bool isUARTEnabled;
        public bool isI2CEnabled;
        public bool isDigitalIOEnabled;
        public bool isUART2Enabled;
        public bool isButtonEnabled;
        public bool isValid;

    }



    /*
     * ------ UART経由で受信したバイトの取り扱い ------
     * USB-UART経由では1バイトずつの転送となるため、1バイトずつ処理
     * (1) 先端フィールドの4バイトが プリアンブル 00 00 AA AA を取得
     * (2) 情報フィールドのデータ長 L を取得
     * (3) 情報フィールドのデータ を受け取った後、
     * (4) 終端フィールドの 4 バイトを受け取る
     * (5) 終端フィールドの末尾2バイトが、ポストアンブル 0D 0A であることを確認
     * (6) 情報フィールドのデータを抜き出し処理にかける
     */

    public class ByteHandler
    {
        public UInt16 num; // 受け取ったバイトの数 
        public byte payload_len; // 情報フィールドの長さ
        public byte payload_type; // データのタイプ ※モーター測定値情報の場合は 0xB4

        public byte payload_num; // 情報フィールドのデータを格納した数
        public byte[] payload; // 情報フィールドのデータを格納する配列
        public byte[] crc_buf; // 付加されたCRCの配列


        public ByteHandler()
        {
            reset();
        }

        public void reset()
        {
            num = 0;
            payload_len = 0;
            payload_type = 0;
            // TODO id

            payload_num = 0;
            payload = new byte[256];
            crc_buf = new byte[2];
        }

    }

    private void HandleDeviceInfo(String infoText)
    {
        Console.WriteLine(infoText);
        string[] arr = infoText.Split(',');

        if (arr.Length >= 4)
        {
            string a = arr[3];
            string[] arr1 = a.Split(':');
            if (arr1.Length >= 2)
            {
                versionText = arr1[1];
                //Console.WriteLine(versionText);
                String str = Regex.Replace(versionText, @"[a-zA-Z]", ""); // アルファベットを除去
                //Console.WriteLine(str);
                versionNumber = float.Parse(str);
                //Console.WriteLine(versionNumber);

            }

        }
    }

    private const byte preamble_len = 4;
    // private const byte postamble_len = 4;

    public ByteHandler h = new ByteHandler();

    public void HandleReceivedByte(byte b)
    {
        // Console.WriteLine("[{0}]: {1}",h.num, b);
        if (h.num == 0 && b == 0x00)
        {
            // プリアンブルの確認: 0x00
            h.num++;
        }
        else if (h.num == 1 && b == 0x00)
        {
            // プリアンブルの確認: 0x00
            h.num++;
        }
        else if (h.num == 2 && b == 0xAA)
        {
            // プリアンブルの確認: 0xAA
            h.num++;
        }
        else if (h.num == 3 && b == 0xAA)
        {
            // プリアンブルの確認: 0xAA
            h.num++;
        }
        else if (h.num == 4)
        {
            // 情報フィールドのデータ長を取得
            h.payload_len = b;
            h.payload[h.payload_num] = b;
            h.payload_num++;
            h.num++;
        }
        else if (h.num > 4 && h.num < preamble_len + h.payload_len)
        {
            // 情報フィールドのデータを格納
            h.payload[h.payload_num] = b;
            h.payload_num++;
            h.num++;
            //Console.WriteLine("byte: {0}", b);
        }
        else if (h.num == preamble_len + h.payload_len)
        {
            // CRC1バイト目（誤り検出）
            h.crc_buf[0] = b;
            h.num++;
        }
        else if (h.num == preamble_len + h.payload_len + 1)
        {
            // CRC2バイト目（誤り検出）
            h.crc_buf[1] = b;
            h.num++;
        }
        else if (h.num == preamble_len + h.payload_len + 2 && b == 0x0D)
        {
            // ポストアンブルの確認: 0x0D
            h.num++;
        }
        else if (h.num == preamble_len + h.payload_len + 3 && b == 0x0A)
        {
            // ポストアンブルの確認: 0x0A
            HandlePayload(h.payload, h.payload_len, h.crc_buf);
            //Console.WriteLine("Payload num and len {0} = {1}", h.payload_len, h.payload_num);
            //Console.WriteLine(string.Join(", ", h.payload));
            //Console.WriteLine(string.Join(", ", h.crc_buf));
            h.reset();
        }
        else
        {
            h.reset();
        }
        
    }


    void HandlePayload(byte[] payload, UInt16 len, byte[] crc_buf)
    {
        byte[] buf = new byte[len];
        Array.Copy(payload, buf, len);
        byte[] array = buf.ToArray();

        byte payload_len = payload[0];
        byte payload_type = payload[1];

        UInt16 crc_payload = CRC16(payload, len);
        UInt16 crc = CRC16(crc_buf, 2);

        
        //Console.WriteLine("CRC Validation: {0} = {1} ", crc_payload, crc);


        // モーター情報
        if (payload_type == 0xB4)
        {
            position = BytesToFloat(array.Skip(2).Take(4).ToArray());
            degree = (float)(position / Math.PI * 180);
            velocity = BytesToFloat(array.Skip(6).Take(4).ToArray());
            rpm = (float)(velocity / (2 * Math.PI) * 60);
            torque = BytesToFloat(array.Skip(10).Take(4).ToArray());

            //Console.WriteLine("pos, vel, trq = {0}, {1}, {2}", position, velocity, torque);

            MotorMeasurementEvent(this, position, velocity, torque);

        }
        // IMU情報
        else if (payload_type == 0xB5)
        {
            accelX = BytesToInt16(array.Skip(2).Take(2).ToArray());
            accelY = BytesToInt16(array.Skip(4).Take(2).ToArray());
            accelZ = BytesToInt16(array.Skip(6).Take(2).ToArray());
            temp = BytesToInt16(array.Skip(8).Take(2).ToArray());
            gyroX = BytesToInt16(array.Skip(10).Take(2).ToArray());
            gyroY = BytesToInt16(array.Skip(12).Take(2).ToArray());
            gyroZ = BytesToInt16(array.Skip(14).Take(2).ToArray());

            IMUMeasurementEvent(this, accelX, accelY, accelZ, temp, gyroX, gyroY, gyroZ);

        }
        // ReadRegisters 
        else if (payload_type == 0x40)
        {

            //Console.WriteLine("Read received: {0}", BitConverter.ToString(array));
            
            byte register = array[4]; // レジスタ（コマンド）
            byte[] valueBytes = array.Skip(5).ToArray(); // 値の配列
            int valueLen = valueBytes.Length; // 値のサイズ
            byte[] idArray = array.Skip(2).Take(2).ToArray();
            UInt16 id = BytesToUInt16(idArray); // ID

            switch (register)
            {
                case 0x07: // ACC
                    acc = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x08: // DEC
                    dec = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x9A: // ReadStatus
                    Status s = decodeStatus(valueBytes);
                    status = s;
                    ReadStatusEvent(this, s);
                    Console.WriteLine("status");
                    break;

                case 0x25: // ReadPIDTable
                    byte tableIndex = array[5];
                    byte paramType = array[6];
                    float val = BytesToFloat(array.Skip(7).Take(4).ToArray());
                    ReadPIDTableEvent(this, id, tableIndex, (PIDParamType)paramType, val);
                    Console.WriteLine("pid");
                    break;
                case 0x18:
                    
                case 0x19:

                case 0x1B:
                    speedP = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x1C:
                    speedI = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x1D:
                    speedD = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x1E:
                    positionP = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x1F:
                    positionI = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x20:
                    positionD = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x21:
                    positionPIDThreshold = BytesToFloat(array.Skip(5).Take(4).ToArray());
                    break;
                case 0x2B:
                    notifyPosArrival = (array[5] == 1) ? true : false;
                    posArrivalTolerance = BytesToFloat(array.Skip(6).Take(4).ToArray());
                    posArrivalJudgeTime = BytesToUInt32(array.Skip(10).Take(4).ToArray());
                    break;

                case 0x2E: // ReadInterface
                    interfaceSetting.rawValue = valueBytes[0];
                    byte rv = interfaceSetting.rawValue;
                    interfaceSetting.isBLEEnabled = ((rv & 1) == 1);
                    interfaceSetting.isLinkageEnabled = (((rv >> 1) & 1) == 1);
                    interfaceSetting.isRadioEnabled = (((rv >> 2) & 1) == 1);
                    interfaceSetting.isUARTEnabled = (((rv >> 3) & 1) == 1);
                    interfaceSetting.isI2CEnabled = (((rv >> 4) & 1) == 1);
                    interfaceSetting.isDigitalIOEnabled = (((rv >> 5) & 1) == 1);
                    interfaceSetting.isUART2Enabled = (((rv >> 6) & 1) == 1);
                    interfaceSetting.isButtonEnabled = (((rv >> 7) & 1) == 1);
                    break;
                case 0x46: // ReadDeviceName
                    int length = valueBytes.Length;
                    Console.WriteLine("len: {0}", length);
                    byte[] data = new byte[length + 1];
                    data[length] = 0;
                    valueBytes.CopyTo(data, 0);
                    deviceName = System.Text.Encoding.ASCII.GetString(valueBytes);
                    break;
                case 0x47: // ReadDeviceInfo
                    int len1 = valueBytes.Length;
                    Console.WriteLine("len: {0}", len1);
                    byte[] data1 = new byte[valueLen + 1];
                    data1[valueLen] = 0;
                    valueBytes.CopyTo(data1, 0);
                    deviceInfo = System.Text.Encoding.ASCII.GetString(valueBytes);
                    HandleDeviceInfo(deviceInfo);
                    break;
                case 0xA6: // ReadTasksetInfo
                    int len2 = valueBytes.Length;
                    tasksetIndex = BytesToUInt16(valueBytes.Take(2).ToArray());
                    tasksetSize = BytesToUInt32(valueBytes.Skip(2).Take(4).ToArray());
                    tasksetName = System.Text.Encoding.UTF8.GetString(valueBytes.Skip(6).Take(15).ToArray());
                    Console.WriteLine("ReadTasksetInfo received index: {0}, size: {1}, name: {2}", tasksetIndex, tasksetSize, tasksetName);
                    break;
                case 0xDB: // ReadRadioSettings
                    UInt32 keyValue = BytesToUInt32(valueBytes.Take(4).ToArray());
                    radioKey = System.Text.Encoding.ASCII.GetString(valueBytes.Take(4).ToArray());

                    //radioKey = valueBytes.;
                    radioGroup = valueBytes[4];
                    radioMode = (RadioMode)valueBytes[5];
                    break;

                default:

                    break;
            } 

        }

        else if (payload_type == 0xB7) // ReadMotionレスポンス
        {
            // | 0 1   | 2 3 4 5            | 6 7 8 9            | 10 11 12 13        | 14... 
            // | index | total_pos_num      | pos_index          | pos_num            | 位置座標の配列 float x0, x1...
            // |       | 記録されている全ての位置座標の総数 n	
            //                              | 本payloadに含まれる最初の座標のインデックス （0~n）		
            //                                                   | 本payloadに含まれる座標の総数（最大50個）		
            //             

            byte[] idArray = array.Skip(2).Take(2).ToArray();
            UInt16 id = BytesToUInt16(idArray); // ID
            UInt32 total_pos_num = BytesToUInt32(array.Skip(4).Take(4).ToArray()); // 記録されている全ての位置座標の総数 n	
            UInt32 pos_index = BytesToUInt32(array.Skip(8).Take(4).ToArray()); // 本payloadに含まれる最初の座標のインデックス （0~n）		
            UInt32 pos_num = BytesToUInt32(array.Skip(12).Take(4).ToArray()); // 本payloadに含まれる座標の総数（最大50個）	
            float[] pos_array = new float[50];
            //Console.WriteLine("total {0}", total_pos_num);

            for (int i = 0; i < pos_num; i++)
            {
                float pos = BytesToFloat(array.Skip(16+i*4).Take(4).ToArray());
                //Console.WriteLine("pos {0}: {1}", pos_index + i, pos);
                pos_array[i] = pos;
            }

            ReadMotionEvent(this, id, total_pos_num, pos_index, pos_num, pos_array);

        }
        else if (payload_type == 0xB8) // ReadTasksetレスポンス
        {
            /*
             *
             * | 0, 1            | 2, 3, 4, 5   | 6, 7     | 8        | 9, 10...
             * | taskset.index   | taskset.size | task_cnt | task_len |タスク（コマンド）データ cmd, id, values...
             */

            byte[] indexArray = array.Skip(2).Take(2).ToArray();
            UInt16 index = BytesToUInt16(indexArray); // ID
            UInt32 taskset_size = BytesToUInt32(array.Skip(4).Take(4).ToArray()); //
            UInt16 task_cnt = BytesToUInt16(array.Skip(8).Take(2).ToArray()); // 本payloadに含まれる最初の座標のインデックス （0~n）		
            byte task_len = array[10];
            //Console.WriteLine("task_len: {0}", task_len);

            byte[] task_array = array.Skip(11).Take(task_len).ToArray(); //new byte[32];
            byte taskCmd = task_array[0];
            UInt16 taskId = BytesToUInt16(task_array.Skip(1).Take(2).ToArray());

            Console.WriteLine("Received Taskset index: {0}, size: {1}, cnt: {2}, len: {3}, cmd: 0x{4}", index, taskset_size, task_cnt, task_len, task_array[0].ToString("x2"));

            ReadTasksetEvent(this, index, taskset_size, task_cnt, task_len, task_array, taskCmd, taskId);

            //Console.WriteLine("array: {0}", array);

        }
        else if (payload_type == 0xBE)
        {
            // Error Handle
            Console.WriteLine(BitConverter.ToString(array));
            // id TODO
            byte cmd = payload[4];
            byte[] idArray = array.Skip(2).Take(2).ToArray();
            UInt16 id = BytesToUInt16(idArray);
            //Console.WriteLine("received ID: {0}", id);
            byte[] errorArray = array.Skip(5).Take(4).ToArray();
            var code = BytesToUInt32(errorArray);
            //Console.WriteLine(BitConverter.ToString(errorArray));
            //Console.WriteLine(code);
            ErrorCode errorCode = (ErrorCode)BytesToUInt32(errorArray);
            byte[] infoArray = array.Skip(9).Take(4).ToArray();


            if (errorCode == ErrorCode.KM_SUCCESS_ARRIVAL)
            {
                // PosArrival
                float arrivePos = BytesToFloat(infoArray);
                Console.WriteLine("[Received Success] Arrival Pos:{0}", arrivePos);
                ReceivedPosArrivalEvent(this, id, arrivePos);
            } else
            {   
                UInt32 info = BytesToUInt32(infoArray);
                if (errorCode == ErrorCode.KM_SUCCESS)
                {
                    // Success
                    Console.WriteLine("[Received Success] Command:{0:X} Info:{1}", cmd, info);

                } else
                {
                    // Error
                    Console.WriteLine("[Received Error] Command: {0:X}, Code:{1}, Info:{2}", cmd, errorCode, info);
                }

                ReceivedErrorEvent(this, id, cmd, errorCode, info);


            }


        }


        //System.Console.ReadLine();

    }

    private Status decodeStatus(byte[] statusBytes)
    {
        byte bs = statusBytes[0];
        Status status = new Status();
        status.isEnabled = ((bs & 1) == 1);
        status.isQueuePaused = (((bs >> 1) & 1) == 1);
        status.isMotorMeasurementEnabled = (((bs >> 2) & 1) == 1);
        status.isIMUMeasurementEnabled = (((bs >> 3) & 1) == 1);
        status.isCheckSumEnabled = (((bs >> 7) & 1) == 1);
        status.flashMemoryState = statusBytes[1];
        status.motorControlMode = statusBytes[2];

        return status;
    }


    //private float BytesToFloat(IEnumerable<byte> enumerable)
    //{
    //  throw new NotImplementedException();
    //}
}
